"""
symbolite.impl.liblatex
~~~~~~~~~~~~~~~~~~~~~

Translate symbolite
into values and functions defined in Python into latex.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from .. import Kind
from . import lang, real

KIND = Kind.CODE
__all__ = ["real", "lang"]



"""
symbolite.ops._substitute
~~~~~~~~~~~~~~~~~~~~~~~~~

Replace symbols, functions, values, etc by others.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from collections.abc import Mapping
from functools import singledispatch
from typing import Any

from ..abstract.vector import getitem
from ..core.call import Call
from ..core.lang import Assign, Block
from ..core.symbolite_object import get_symbolite_info
from ..core.value import Name, Value


@singledispatch
def substitute(obj: Any, replacements: Mapping[Any, Any]) -> Any:
    """Replace symbols, functions, values, etc by others.

    Parameters
    ----------
    obj
        symbolic expression.
    replacements
        replacement dictionary.
    """
    return replacements.get(obj, obj)


@substitute.register(Value)
def substitute_value[R: Value[Any]](obj: R, mapper: Mapping[Any, Any]) -> R:
    info = get_symbolite_info(obj)
    if isinstance(info.value, Name):
        return mapper.get(obj, obj)
    return obj.__class__(substitute(info.value, mapper))


@substitute.register
def substitue_call(obj: Call, mapper: Mapping[Any, Any]) -> Call:
    info = get_symbolite_info(obj)
    func = mapper.get(info.func, info.func)
    args = tuple(substitute(arg, mapper) for arg in info.args)
    kwargs = {k: substitute(arg, mapper) for k, arg in info.kwargs_items}

    return Call(func, args, tuple(kwargs.items()))


@substitute.register
def substitue_assign(obj: Assign, mapper: Mapping[Any, Any]) -> Assign:
    info = get_symbolite_info(obj)
    return Assign(
        substitute(info.lhs, mapper),
        substitute(info.rhs, mapper),
    )


def is_vector_item(el: Value[Any]) -> tuple[bool, tuple[Any, ...]]:
    info = get_symbolite_info(el)
    if not isinstance(info.value, Call):
        return False, ()
    call = info.value
    info = get_symbolite_info(call)
    if info.func is not getitem:
        return False, ()
    return True, info.args


def _subs(
    content: tuple[Value[Any], ...], mapper: Mapping[Any, Any]
) -> tuple[tuple[Value[Any], ...], tuple[Value[Any], ...]]:
    out: list[Value[Any]] = []
    vec_added: list[Value[Any]] = []
    for el in content:
        if el not in mapper:
            out.append(el)
            continue
        mapped = mapper[el]
        sel, args = is_vector_item(mapped)
        if sel:
            var, _ndx = args
            if var not in out:
                out.append(var)
                vec_added.append(var)
        else:
            out.append(mapped)

    return tuple(out), tuple(vec_added)


@substitute.register
def substitue_block(obj: Block, mapper: Mapping[Any, Any]) -> Block:
    info = get_symbolite_info(obj)

    inputs, _ = _subs(info.inputs, mapper)
    outputs, out_vec = _subs(info.outputs, mapper)

    return Block(
        inputs + out_vec,
        outputs,
        lines=tuple(substitute(el, mapper) for el in info.lines),
        name=info.name,
    )



from ...abstract import real as abstract_real
from ...core.function import (
    BinaryFunction,
    BinaryOperator,
    UnaryFunction,
    UnaryOperator,
)

latex_names = {
    abstract_real.eq: "{} = {}",
    abstract_real.ne: "{} \\neq {}",
    abstract_real.lt: "{} < {}",
    abstract_real.le: "{} \\le {}",
    abstract_real.gt: "{} > {}",
    abstract_real.ge: "{} \\ge {}",
    abstract_real.add: "{} + {}",
    abstract_real.sub: "{} - {}",
    abstract_real.mul: "{} \\cdot {}",
    abstract_real.truediv: "\\dfrac{{{}}}{{{}}}",
    abstract_real.mod: "{} \\pmod{{{}}}",
    abstract_real.lshift: "{} \\ll {}",
    abstract_real.rshift: "{} \\gg {}",
    abstract_real.and_: "{} \\land {}",
    abstract_real.xor: "{} \\oplus {}",
    abstract_real.or_: "{} \\lor {}",
    abstract_real.neg: "-{}",
    abstract_real.pos: "+{}",
    abstract_real.invert: "~{}",
    abstract_real.abs: "\\left\\lvert {} \\right\\rvert",
    abstract_real.acos: "\\arccos",
    abstract_real.acosh: "\\operatorname{arcosh}",
    abstract_real.asin: "\\arcsin",
    abstract_real.asinh: "\\operatorname{arcsinh}",
    abstract_real.atan: "\\arctan",
    abstract_real.atan2: "\\operatorname{arctan2}",
    abstract_real.atanh: "\\operatorname{arctanh}",
    abstract_real.ceil: "\\left\\lceil {} \\right\\rceil",
    abstract_real.comb: "\\binom{n}{k}",
    abstract_real.copysign: "\\operatorname{copysign}",
    abstract_real.cos: "\\cos",
    abstract_real.cosh: "\\cosh",
    abstract_real.degrees: "\\operatorname{deg}",
    abstract_real.erf: "\\operatorname{erf}",
    abstract_real.erfc: "\\operatorname{erfc}",
    abstract_real.exp: "\\exp",
    abstract_real.expm1: "\\operatorname{expm1}",
    abstract_real.factorial: "{}!",
    abstract_real.floor: "\\left\\lfloor {} \\right\\rfloor",
    abstract_real.fmod: "\\operatorname{fmod}",
    abstract_real.frexp: "\\operatorname{frexp}",
    abstract_real.gamma: "\\Gamma",
    abstract_real.hypot: "\\operatorname{hypot}",
    abstract_real.isfinite: "\\operatorname{isfinite}",
    abstract_real.isinf: "\\operatorname{isinf}",
    abstract_real.isnan: "\\operatorname{isnan}",
    abstract_real.ldexp: "\\operatorname{ldexp}",
    abstract_real.lgamma: "\\ln\\Gamma",
    abstract_real.log: "\\ln",
    abstract_real.log10: "\\log_{10}",
    abstract_real.log1p: "\\operatorname{log1p}",
    abstract_real.log2: "\\log_2",
    abstract_real.modf: "\\operatorname{modf}",
    abstract_real.nextafter: "\\operatorname{nextafter}",
    abstract_real.radians: "\\operatorname{rad}",
    abstract_real.remainder: "\\operatorname{rem}",
    abstract_real.sin: "\\sin",
    abstract_real.sinh: "\\sinh",
    abstract_real.sqrt: "\\sqrt{{{}}}",
    abstract_real.tan: "\\tan",
    abstract_real.tanh: "\\tanh",
    abstract_real.trunc: "\\operatorname{trunc}",
    abstract_real.ulp: "\\operatorname{ulp}",
    abstract_real.le: "{} \\le {}",
    abstract_real.pow: "{}^{{{}}}",
    abstract_real.e: "e",
    abstract_real.inf: "\\infty",
    abstract_real.pi: "\\pi",
    abstract_real.nan: "\\text{NaN}",
    abstract_real.tau: "\\tau",
}


def _get_latex_name(
    obj: UnaryFunction | BinaryFunction | UnaryOperator | BinaryOperator,
) -> str:
    return latex_names[obj]


latex_precedences = {
    abstract_real.abs: 5,
    abstract_real.ceil: 5,
    abstract_real.floor: 5,
    abstract_real.factorial: 4,
    abstract_real.pow: 3,
    abstract_real.sqrt: 3,
}
# Since division is implemented with \dfrac parenthesis are never necessary,
# so it is given the lowest precedence


def _get_latex_precedence(
    obj: UnaryFunction | BinaryFunction | UnaryOperator | BinaryOperator,
) -> int:
    """Some functions are turned into operators for latex conversion
    or given a different precedence."""
    precedence = latex_precedences.get(
        obj, getattr(obj.__symbolite_info__, "precedence", None)
    )
    if precedence is None:
        raise AttributeError(f"{obj} has no precedence")
    else:
        return precedence


never_parentesize = [
    abstract_real.abs,
    abstract_real.floor,
    abstract_real.ceil,
    abstract_real.truediv,
]


def _parenthesize(
    obj: UnaryFunction | BinaryFunction | UnaryOperator | BinaryOperator,
) -> bool:
    return obj not in never_parentesize



from __future__ import annotations

from typing import Any

from ...core import Unsupported
from ...ops._translate import translate

Block = Unsupported

Assign = Unsupported


def to_bool(value: bool, libsl: Any) -> str:
    return "True" if value else "False"


def to_int(value: int, libsl: Any) -> str:
    return repr(value)


def to_float(value: float, libsl: Any) -> str:
    return repr(value)


def to_tuple(value: tuple[Any, ...], libsl: Any) -> str:
    value = (translate(v, libsl) for v in value)
    return f"({', '.join(map(str, value))}, )"


def to_list(value: tuple[Any, ...], libsl: Any) -> str:
    value = (translate(v, libsl) for v in value)
    return f"[{', '.join(map(str, value))}]"


def to_dict(value: tuple[tuple[Any, Any], ...], libsl: Any) -> str:
    items = to_tuple(value, libsl)
    return f"dict({items})"


__all__ = [
    "Block",
    "Assign",
    "to_bool",
    "to_int",
    "to_float",
    "to_tuple",
    "to_list",
    "to_dict",
]



"""
symbolite.impl.libpythoncode
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Utilities to build Python source snippets from Symbolite expressions.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

from collections.abc import Iterable
from dataclasses import dataclass
from typing import Any

from ...core.function import Function, Operator
from ...core.symbolite_object import get_symbolite_info
from ...core.value import Name, Value
from ._latex_names import _get_latex_name, _get_latex_precedence, _parenthesize


@dataclass(frozen=True)
class LatexExpr:
    """Represents a snippet of Python code plus its precedence."""

    text: str
    precedence: int = 100

    def __str__(self) -> str:  # pragma: no cover - convenience
        return self.text


def _coerce(value: Any) -> LatexExpr:
    if isinstance(value, LatexExpr):
        return value
    if isinstance(value, str):
        return LatexExpr(value)
    if isinstance(value, bool):
        return LatexExpr("True" if value else "False")
    if isinstance(value, (int, float, complex)):
        return LatexExpr(repr(value))
    return LatexExpr(str(value))


def _join_arguments(
    args: Iterable[LatexExpr], kwargs: Iterable[tuple[str, LatexExpr]]
) -> str:
    parts = [arg.text for arg in args]
    parts.extend(f"{key}={value.text}" for key, value in kwargs)
    return ", ".join(parts)


def _maybe_parenthesize(
    expr: LatexExpr, precedence: int, *, right: bool, parenthesize: bool
) -> str:
    if expr.precedence < precedence and parenthesize:
        return f"\\left({expr.text}\\right)"
    if right and expr.precedence == precedence and parenthesize:
        return f"\\left({expr.text}\\right)"
    return expr.text


def make_function(latex_name: str) -> Any:
    def _function(*args: Any, **kwargs: Any) -> LatexExpr:
        coerced_args = tuple(_coerce(arg) for arg in args)
        coerced_kwargs = tuple((k, _coerce(v)) for k, v in kwargs.items())
        return LatexExpr(
            f"{latex_name}\\left( {_join_arguments(coerced_args, coerced_kwargs)} \\right)"
        )

    return _function


def make_operator(fmt: str, precedence: int, arity: int, parenthesize: bool) -> Any:
    def _operator(*args: Any) -> LatexExpr:
        coerced = tuple(_coerce(arg) for arg in args)
        if arity == 1:
            (value,) = coerced
            formatted = fmt.format(
                _maybe_parenthesize(
                    value,
                    precedence,
                    right=False,
                    parenthesize=parenthesize,
                )
            )
        else:
            leading = coerced[0]
            trailing = coerced[1:]
            formatted_args = [
                _maybe_parenthesize(
                    leading,
                    precedence,
                    right=False,
                    parenthesize=parenthesize,
                ),
                *(
                    _maybe_parenthesize(
                        arg,
                        precedence,
                        right=index == len(trailing) - 1,
                        parenthesize=parenthesize,
                    )
                    for index, arg in enumerate(trailing)
                ),
            ]
            formatted = fmt.format(*formatted_args)
        return LatexExpr(formatted, precedence)

    return _operator


def make_attribute(name: str) -> LatexExpr:
    return LatexExpr(name)


def as_function(obj: Function[Any]) -> Any:
    latex_name = _get_latex_name(obj)
    return make_function(latex_name)


def as_operator(obj: Operator[Any]) -> Any:
    info = get_symbolite_info(obj)
    return make_operator(
        _get_latex_name(obj),
        _get_latex_precedence(obj),
        info.arity,
        parenthesize=_parenthesize(obj),
    )


def as_named_value(obj: Value[Any]) -> LatexExpr:
    info = get_symbolite_info(obj)
    if not isinstance(info.value, Name):
        raise ValueError(f"Value {obj!r} is not bound to a Name.")
    latex_name = _get_latex_name(obj)
    return make_attribute(latex_name)


def _nest_with_brackets(text: str, separator: str) -> str:
    parts = text.split(separator)

    result = parts[-1]

    for part in reversed(parts[:-1]):
        result = f"{part}{separator}{{{result}}}"

    return result


def nest_sub_and_super_scripts(text: str) -> str:
    return _nest_with_brackets(_nest_with_brackets(text, "_"), "^")



from typing import Callable

import pytest

from ..abstract import real as abstract_real
from ..impl import liblatex
from ..impl.liblatex._latexexpr import LatexExpr

x, y, z = map(abstract_real.Real, "x y z".split())
dummy_reals = (x, y, z)

all_funcs = {impl: getattr(liblatex.real, impl) for impl in liblatex.real.__all__}


@pytest.mark.parametrize("func", all_funcs.values(), ids=all_funcs.keys())
def test_all_real_funcs_run(func: Callable | LatexExpr):
    if func == liblatex.real.Real:
        assert isinstance(func("x"), LatexExpr)
        return
    # TODO: set what each function should be tested as
    try:
        # Test as unary function
        func(*dummy_reals[:1])
    except IndexError:
        # Test as binary function
        func(*dummy_reals[:2])
    except TypeError:
        # Test as named value
        assert isinstance(func, LatexExpr)
